from gcp_pilot.mocker import patch_auth


class ClientTestMixin:
    _CLIENT_KLASS = None

    def get_client(self, **kwargs):
        return self._CLIENT_KLASS(**kwargs)

    def test_mocked_auth(self):
        with patch_auth():
            self.get_client()



# More Information <https://cloud.google.com/scheduler/docs/reference/rest>
import os
from collections.abc import Generator

from google.api_core.exceptions import NotFound
from google.cloud import scheduler

from gcp_pilot.base import AppEngineBasedService, GoogleCloudPilotAPI

DEFAULT_TIMEZONE = os.environ.get("TIMEZONE", "Etc/UTC")  # UTC
MAX_TIMEOUT = 30 * 60  # max allowed to HTTP endpoints is 30 minutes


class CloudScheduler(AppEngineBasedService, GoogleCloudPilotAPI):
    _client_class = scheduler.CloudSchedulerClient
    DEFAULT_METHOD = scheduler.HttpMethod.POST

    def __init__(self, **kwargs):
        self.timezone = kwargs.pop("timezone", DEFAULT_TIMEZONE)
        super().__init__(**kwargs)

    def _parent_path(self, project_id: str | None = None) -> str:
        return self._location_path(project_id=project_id, location=self.location)

    def _job_path(self, job, project_id: str | None = None) -> str:
        parent_name = self._parent_path(project_id=project_id)
        return f"{parent_name}/jobs/{job}"

    def create(
        self,
        name: str,
        url: str,
        payload: str,
        cron: str,
        timezone: str = DEFAULT_TIMEZONE,
        method: int = DEFAULT_METHOD,
        headers: dict[str, str] | None = None,
        project_id: str | None = None,
        use_oidc_auth: bool = True,
        timeout_in_seconds: int = MAX_TIMEOUT,
        retry_count: int = 0,
    ) -> scheduler.Job:
        parent = self._parent_path(project_id=project_id)
        job_name = self._job_path(job=name, project_id=project_id)
        job = scheduler.Job(
            name=job_name,
            schedule=cron,
            time_zone=timezone or self.timezone,
            attempt_deadline=self._as_duration(seconds=timeout_in_seconds),
            http_target=scheduler.HttpTarget(
                uri=url,
                http_method=method,
                body=payload.encode(),
                headers=headers or {},
                **(self.get_oidc_token(audience=url) if use_oidc_auth else {}),
            ),
            retry_config=scheduler.RetryConfig(
                retry_count=retry_count,
            ),
        )

        response = self.client.create_job(request={"parent": parent, "job": job})
        return response

    def update(
        self,
        name: str,
        url: str,
        payload: str,
        cron: str,
        timezone: str | None = DEFAULT_TIMEZONE,
        method: int = DEFAULT_METHOD,
        headers: dict[str, str] | None = None,
        project_id: str | None = None,
        use_oidc_auth: bool = True,
        timeout_in_seconds: int = MAX_TIMEOUT,
        retry_count: int = 0,
    ) -> scheduler.Job:
        job_name = self._job_path(job=name, project_id=project_id)
        job = scheduler.Job(
            name=job_name,
            schedule=cron,
            time_zone=timezone or self.timezone,
            attempt_deadline=self._as_duration(seconds=timeout_in_seconds),
            http_target=scheduler.HttpTarget(
                uri=url,
                http_method=method,
                body=payload.encode(),
                headers=headers or {},
                **(self.get_oidc_token(audience=url) if use_oidc_auth else {}),
            ),
            retry_config=scheduler.RetryConfig(
                retry_count=retry_count,
            ),
        )

        response = self.client.update_job(
            job=job,
        )
        return response

    def list(self, prefix: str = "", project_id: str | None = None) -> Generator[scheduler.Job]:
        parent = self._parent_path(project_id=project_id)
        for job in self.client.list_jobs(parent=parent):
            if job.name.split("/jobs/")[-1].startswith(prefix):
                yield job

    def get(self, name: str, project_id: str | None = None) -> scheduler.Job:
        job_name = self._job_path(job=name, project_id=project_id)
        return self.client.get_job(
            name=job_name,
        )

    def delete(self, name: str, project_id: str | None = None) -> None:
        job_name = self._job_path(job=name, project_id=project_id)
        return self.client.delete_job(
            name=job_name,
        )

    def put(
        self,
        name: str,
        url: str,
        payload: str,
        cron: str,
        timezone: str = DEFAULT_TIMEZONE,
        method: int = DEFAULT_METHOD,
        headers: dict[str, str] | None = None,
        project_id: str | None = None,
        use_oidc_auth: bool = True,
        timeout_in_seconds: int = MAX_TIMEOUT,
        retry_count: int = 0,
    ) -> scheduler.Job:
        try:
            response = self.update(
                name=name,
                url=url,
                payload=payload,
                cron=cron,
                timezone=timezone,
                method=method,
                headers=headers,
                project_id=project_id,
                use_oidc_auth=use_oidc_auth,
                timeout_in_seconds=timeout_in_seconds,
                retry_count=retry_count,
            )
        except NotFound:
            response = self.create(
                name=name,
                url=url,
                payload=payload,
                cron=cron,
                timezone=timezone,
                method=method,
                headers=headers,
                project_id=project_id,
                use_oidc_auth=use_oidc_auth,
                timeout_in_seconds=timeout_in_seconds,
                retry_count=retry_count,
            )
        return response


__all__ = ("CloudScheduler",)



import unittest

from gcp_pilot.source import SourceRepository
from tests import ClientTestMixin


class TestSourceRepository(ClientTestMixin, unittest.TestCase):
    _CLIENT_KLASS = SourceRepository



import json
import unittest
from datetime import UTC, datetime
from pathlib import Path

from gcp_pilot.factories.identity_platform import FirebaseAuthTokenFactory
from gcp_pilot.identity_platform import FirebaseAuthToken, IdentityPlatform, parse_timestamp
from gcp_pilot.mocker import patch_firebase_token
from tests import ClientTestMixin


class TestParseTimestamp(unittest.TestCase):
    def test_parse_timestamp(self):
        sometime = datetime(2022, 5, 17, 1, 2, 57, tzinfo=UTC)
        sometimestamp = int(sometime.timestamp())
        generated_time = parse_timestamp(sometimestamp)
        self.assertEqual(sometime, generated_time)


class TestIdentityPlatform(ClientTestMixin, unittest.TestCase):
    _CLIENT_KLASS = IdentityPlatform

    @classmethod
    def _load_sample_token(cls, sample_name: str) -> tuple[dict, FirebaseAuthToken]:
        sample_path = Path(__file__).parent / "samples" / "identity_platform" / f"{sample_name}.json"
        token_data = json.load(sample_path.open())
        with patch_firebase_token(return_value=token_data):
            return token_data, FirebaseAuthToken(jwt_token="potato")

    def assert_expected_sample_token(
        self,
        expected_data: dict,
        token: FirebaseAuthToken,
        is_tenant: bool,
        is_expired: bool = True,
    ):
        self.assertEqual(parse_timestamp(expected_data["exp"]), token.expiration_date)
        self.assertEqual(expected_data["event_type"], token.event_type)
        self.assertEqual(expected_data["event_id"], token.event_id)
        self.assertEqual(expected_data["ip_address"], token.ip_address)
        self.assertEqual(expected_data["sign_in_method"], token.provider_id)
        self.assertEqual(expected_data["user_agent"], token.user_agent)

        self.assertEqual(json.loads(expected_data["raw_user_info"]), token.raw_user)

        self.assertEqual(expected_data["user_record"]["uid"], token.user.id)
        self.assertEqual(expected_data["user_record"].get("email"), token.user.email)
        self.assertEqual(expected_data["user_record"].get("display_name"), token.user.name)
        self.assertEqual(expected_data["user_record"].get("photo_url"), token.user.photo_url)
        self.assertEqual(expected_data["user_record"].get("email_verified"), token.user.verified)
        self.assertEqual(expected_data["user_record"].get("disabled"), token.user.disabled)
        self.assertEqual(
            parse_timestamp(expected_data["user_record"]["metadata"]["creation_time"]),
            token.user.created_at,
        )
        self.assertEqual(
            parse_timestamp(expected_data["user_record"]["metadata"]["last_sign_in_time"]),
            token.user.last_login_at,
        )

        self.assertEqual(expected_data["oauth_access_token"], token.oauth.access_token)
        self.assertEqual(expected_data["oauth_id_token"], token.oauth.id_token)
        self.assertEqual(expected_data["oauth_refresh_token"], token.oauth.refresh_token)
        self.assertEqual(expected_data["oauth_token_secret"], token.oauth.token_secret)

        self.assertEqual(expected_data["aud"], token.jwt_info.aud)
        self.assertEqual(expected_data["iss"], token.jwt_info.iss)
        self.assertEqual(parse_timestamp(expected_data["iat"]), token.jwt_info.iat)
        self.assertEqual(parse_timestamp(expected_data["exp"]), token.jwt_info.exp)
        self.assertEqual(expected_data["sub"], token.jwt_info.sub)
        self.assertEqual(is_expired, token.jwt_info.is_expired)

        if is_tenant:
            self.assertEqual(expected_data["tenant_id"], token.tenant_id)
            self.assertEqual(expected_data["user_record"]["tenant_id"], token.user.tenant_id)
        else:
            self.assertFalse("tenant_id" in expected_data)
            self.assertIsNone(token.tenant_id)
            self.assertFalse("tenant_id" in expected_data["user_record"])
            self.assertIsNone(token.user.tenant_id)

    def test_parse_tenant_oidc_signin_token(self):
        token_data, token = self._load_sample_token("firebase_tenant_oidc_signin_decoded_token")
        self.assert_expected_sample_token(expected_data=token_data, token=token, is_tenant=True)

    def test_parse_tenant_signin_token(self):
        token_data, token = self._load_sample_token("firebase_tenant_signin_decoded_token")
        self.assert_expected_sample_token(expected_data=token_data, token=token, is_tenant=True)

    def test_parse_signin_token(self):
        token_data, token = self._load_sample_token("firebase_signin_decoded_token")
        self.assert_expected_sample_token(expected_data=token_data, token=token, is_tenant=False)

    def test_parse_firebase_factory_token(self):
        iat = int(datetime.now(tz=UTC).timestamp())
        token_data = FirebaseAuthTokenFactory.create(iat=iat)
        with patch_firebase_token(return_value=token_data):
            token = FirebaseAuthToken(jwt_token="potato")

        self.assert_expected_sample_token(expected_data=token_data, token=token, is_tenant=True, is_expired=False)



# More Information <https://cloud.google.com/run/docs/reference/rest>
from collections.abc import Generator

from google.api_core.client_options import ClientOptions
from googleapiclient.discovery import Resource

from gcp_pilot import exceptions
from gcp_pilot.base import DiscoveryMixin, GoogleCloudPilotAPI, ResourceType
from gcp_pilot.resource import ServiceAgent


class CloudRun(DiscoveryMixin, GoogleCloudPilotAPI):
    def _service_endpoint(self, location: str | None = None):
        if not location:
            return "https://run.googleapis.com"
        return f"https://{location}-run.googleapis.com"

    def _namespace_path(self, project_id: str | None = None):
        return f"namespaces/{project_id or self.project_id}"

    def _service_path(self, service_name: str, project_id: str | None = None):
        parent = self._namespace_path(project_id=project_id)
        return f"{parent}/services/{service_name}"

    def _domain_mapping_path(self, domain: str, project_id: str | None = None):
        parent = self._namespace_path(project_id=project_id)
        return f"{parent}/domainmappings/{domain}"

    def list_services(self, project_id: str | None = None) -> Generator[ResourceType]:
        params = dict(
            parent=self._namespace_path(project_id=project_id),
        )
        yield from self._list(
            method=self.client.namespaces().services().list,
            result_key="items",
            params=params,
        )

    def get_service(
        self,
        service_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> ResourceType:
        name = self._service_path(service_name=service_name, project_id=project_id)
        client = self._get_localized_client(project_id=project_id, location=location)
        return self._execute(
            method=client.namespaces().services().get,
            name=name,
        )

    def create_service(
        self,
        service_name: str,
        project_id: str | None = None,
        location: str | None = None,
        service_account: str | None = None,
        trigger_id: str | None = None,
        image: str = "gcr.io/cloudrun/placeholder",
        ram: int = 256,
        concurrency: int = 80,
        timeout: int = 300,
        port: int = 8080,
    ) -> ResourceType:
        parent = self._namespace_path(project_id=project_id)
        client = self._get_localized_client(project_id=project_id, location=location)

        service_account = service_account or ServiceAgent.get_compute_service_account(
            project_id=project_id or self.project_id,
        )
        labels = {
            "managed-by": "gcp-cloud-build-deploy-cloud-run",
            "cloud.googleapis.com/location": location,
        }
        if trigger_id:
            labels["gcb-trigger-id"] = trigger_id

        body = {
            "apiVersion": "serving.knative.dev/v1",
            "kind": "Service",
            "metadata": {
                "name": service_name,
                "labels": labels,
                "annotations": {
                    "client.knative.dev/user-image": image,
                },
            },
            "spec": {
                "template": {
                    "spec": {
                        "containerConcurrency": concurrency,
                        "timeoutSeconds": timeout,
                        "serviceAccountName": service_account,
                        "containers": [
                            {
                                "image": image,
                                "resources": {"limits": {"cpu": "1000m", "memory": f"{ram}Mi"}},
                                "ports": [{"containerPort": port}],
                            },
                        ],
                    },
                },
            },
        }
        return self._execute(
            method=client.namespaces().services().create,
            parent=parent,
            body=body,
        )

    def list_locations(self, project_id: str | None = None) -> Generator[ResourceType]:
        params = dict(
            name=self._project_path(project_id=project_id),
        )
        yield from self._paginate(
            method=self.client.projects().locations().list,
            result_key="locations",
            params=params,
        )

    def list_revisions(
        self,
        service_name: str | None = None,
        project_id: str | None = None,
    ) -> Generator[ResourceType]:
        params = dict(
            parent=self._namespace_path(project_id=project_id),
        )
        items = self._list(
            method=self.client.namespaces().revisions().list,
            result_key="items",
            params=params,
        )
        for item in items:
            if not service_name or item["metadata"]["labels"]["serving.knative.dev/service"] == service_name:
                yield item

    def list_domain_mappings(self, project_id: str | None = None) -> Generator[ResourceType]:
        params = dict(
            parent=self._namespace_path(project_id=project_id),
        )
        yield from self._list(
            method=self.client.namespaces().domainmappings().list,
            result_key="items",
            params=params,
        )

    def get_domain_mapping(
        self,
        domain: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> ResourceType:
        name = self._domain_mapping_path(domain=domain, project_id=project_id)
        client = self._get_localized_client(project_id=project_id, location=location)
        return self._execute(
            method=client.namespaces().domainmappings().get,
            name=name,
        )

    def delete_domain_mapping(
        self,
        domain: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> ResourceType:
        name = self._domain_mapping_path(domain=domain, project_id=project_id)
        client = self._get_localized_client(project_id=project_id, location=location)
        return self._execute(
            method=client.namespaces().domainmappings().delete,
            name=name,
        )

    def create_domain_mapping(
        self,
        domain: str,
        service_name: str,
        project_id: str | None = None,
        location: str | None = None,
        exists_ok: bool = True,
        force: bool = True,
    ) -> ResourceType:
        parent = self._namespace_path(project_id=project_id)
        client = self._get_localized_client(project_id=project_id, location=location)

        body = {
            "apiVersion": "domains.cloudrun.com/v1",
            "kind": "DomainMapping",
            "metadata": {"name": domain},
            "spec": {"routeName": service_name, "certificateMode": "AUTOMATIC", "forceOverride": force},
        }
        try:
            return self._execute(
                method=client.namespaces().domainmappings().create,
                parent=parent,
                body=body,
            )
        except exceptions.AlreadyExists:
            if not exists_ok:
                raise
            return self.get_domain_mapping(
                domain=domain,
                project_id=project_id,
                location=location,
            )

    def _get_localized_client(self, project_id: str | None = None, location: str | None = None) -> Resource:
        # Reminder: List methods do not require a localized client
        if not location:
            location = self.location if not project_id else self._get_project_default_location(project_id=project_id)
        return self._build_client(location=location)

    def _build_client(self, location: str | None = None, **kwargs) -> Resource:
        options = ClientOptions(api_endpoint=self._service_endpoint(location=location))
        kwargs.update(
            dict(
                serviceName="run",
                version="v1",
                cache_discovery=False,
                client_options=options,
            ),
        )
        return super()._build_client(**kwargs)


__all__ = ("CloudRun",)
